@@ -9,7 +9,9 @@ import time
 import yaml
 from astropy.time import TimeDelta
 from glob import glob
-from pyoperators import DenseOperator, HomothetyOperator, Rotation3dOperator
+from pyoperators import (
+    DenseBlockDiagonalOperator, HomothetyOperator, IdentityOperator,
+    Rotation2dOperator, Rotation3dOperator)
 from pyoperators.utils import ifirst
 from pysimulators import (
     Acquisition, CartesianEquatorial2HorizontalOperator,
@@ -17,7 +19,6 @@ from pysimulators import (
     CartesianHorizontal2EquatorialOperator,
     CartesianEquatorial2GalacticOperator)
 from pysimulators.interfaces.healpy import HealpixConvolutionGaussianOperator
-
 from .calibration import QubicCalibration
 from .instrument import QubicInstrument
 
@@ -53,7 +54,7 @@ class QubicAcquisition(Acquisition):
         if not isinstance(instrument, (QubicInstrument, str)):
             raise TypeError("Invalid type for the instrument ('{}' instead of "
                             "'QubicInstrument' or 'str').".format(type(
-                            instrument).__name__))
+                                instrument).__name__))
         if isinstance(instrument, str):
             raise NotImplementedError('Module names not fixed yet.')
         Acquisition.__init__(self, instrument, pointing, block_id=block_id,
@@ -83,10 +84,10 @@ class QubicAcquisition(Acquisition):
         p = self.pointing
         time = p.date_obs + TimeDelta(p.time, format='sec')
         r = CartesianEquatorial2GalacticOperator() * \
+            CartesianHorizontal2EquatorialOperator(
+                'NE', time, p.latitude, p.longitude) * \
             Rotation3dOperator("ZY'Z''", p.azimuth, 90 - p.elevation, p.pitch,
-                               degrees=True) * \
-            CartesianHorizontal2EquatorialOperator('NE', time, p.latitude,
-                                                   p.longitude)
+                               degrees=True)
         return r
 
     def get_rotation_g2i(self):
@@ -98,8 +99,8 @@ class QubicAcquisition(Acquisition):
         time = p.date_obs + TimeDelta(p.time, format='sec')
         r = Rotation3dOperator("ZY'Z''", p.azimuth, 90 - p.elevation, p.pitch,
                                degrees=True).T * \
-            CartesianEquatorial2HorizontalOperator('NE', time, p.latitude,
-                                                   p.longitude) * \
+            CartesianEquatorial2HorizontalOperator(
+                'NE', time, p.latitude, p.longitude) * \
             CartesianGalactic2EquatorialOperator()
         return r
 
@@ -118,11 +119,26 @@ class QubicAcquisition(Acquisition):
         """
         return HealpixConvolutionGaussianOperator(fwhm=fwhm, **keywords)
 
+    def get_detector_response_operator(self, tau=0.01):
+        """
+        Return the operator for the bolometer responses.
+
+        """
+        sampling_period = self.pointing.get_sampling_period()
+        shapein = (self.get_ndetectors, self.pointing.size)
+        return ConvolutionTruncatedExponentialOperator(tau / sampling_period,
+                                                       shapein=shapein)
+
     def get_hwp_operator(self):
         """
         Return the rotation matrix for the half-wave plate.
 
         """
+        if self.instrument.sky.kind == 'I':
+            return IdentityOperator()
+        if self.instrument.sky.kind == 'QU':
+            return Rotation2dOperator(-4 * self.pointing.angle_hwp,
+                                      degrees=True)
         return Rotation3dOperator('X', -4 * self.pointing.angle_hwp,
                                   degrees=True)
 
@@ -131,48 +147,40 @@ class QubicAcquisition(Acquisition):
         Return operator for the polarizer grid.
 
         """
-        if not self.instrument.sky.polarized:
+        if self.instrument.sky.kind == 'I':
             return HomothetyOperator(1 / self.instrument.detector.ngrids)
 
         if self.instrument.detector.ngrids == 1:
-            return DenseOperator([[0.5, 0.5, 0],
-                                  [0.5,-0.5, 0]])
+            raise ValueError(
+                'Polarized input not handled by a single detector grid.')
 
         grid = self.instrument.detector.packed.quadrant // 4
         z = np.zeros(self.get_ndetectors())
-        data = np.array([z + 0.5, 0.5 - grid, z]).T[:, None, :]
+        if self.instrument.sky.kind == 'QU':
+            data = np.array([0.5 - grid, z]).T[:, None, None, :]
+        else:
+            data = np.array([z + 0.5, 0.5 - grid, z]).T[:, None, None, :]
+        return DenseBlockDiagonalOperator(data)
 
-        def direct(x, y):
-            np.sum(x * data, axis=-1, out=y)
-
-        def transpose(x, y):
-            np.multiply(x[..., None], data, out=y)
-
-        shapeout = self.get_ndetectors(), self.get_nsamples()[0]
-        shapein = shapeout + (3,)
-
-        from pyoperators import Operator
-        return Operator(direct=direct, transpose=transpose, dtype=float,
-                        shapein=shapein, shapeout=shapeout,
-                        flags='real, linear')
-
-    def get_projection_peak_operator(self, kmax=2, rotation=None):
+    def get_projection_peak_operator(self, rotation=None, dtype=None,
+                                     synthbeam_fraction=None):
         """
         Return the projection operator for the peak sampling.
 
         Parameters
         ----------
-        kmax : int, optional
-            The diffraction order above which the peaks are ignored.
-            For a value of 0, only the central peak is sampled.
         rotation : array (npointings, 3, 3)
-            The Instrument-to-Reference rotation matrix
+            The Instrument-to-Reference rotation matrix.
+        dtype : dtype
+            The datatype of the elements in the projection matrix.
+        synthbeam_fraction : float, optional
+            Override the instrument synthetic beam fraction.
 
         """
         if rotation is None:
             rotation = self.get_rotation_g2i()
-        return self.instrument.get_projection_peak_operator(rotation,
-                                                            kmax=kmax)
+        return self.instrument.get_projection_peak_operator(
+            rotation, synthbeam_fraction=synthbeam_fraction, dtype=dtype)
 
     @classmethod
     def load(cls, filename, instrument=None, selection=None):
